program mkd(input,output)

{This program generates a distribution
which corresponds to
           random/random (rr)
           eq.random/eq.random (ee)
           grid/grid (gg)
           block/grid (bg)
           diag (dd)
or other combinations
of the report R. H. Bisseling and W. F. McColl,
Scientific Computing on Bulk Synchronous Parallel Architectures,
Preprint 836, Department of Mathematics, Utrecht University,
dec. 1993.

Usage: mkd inpfile distrfile

  This program creates a Cartesian distribution
  of a vector of length n
  over a q0 by q1 processor network,
  using one of the matrix-independent distribution schemes.

 Input: two filenames as arguments
        inpfile   = file containing
                    line 0              one option from x,y,d,g,b,r,e
                    line 1              one option from x,y,d,g,b,r,e
                    line 2              n = vector length
                    line 3              q0 = number of processor rows
                    line 4              q1 = number of processor columns
        distrfile = new distribution file to be created 
Sample input file:
b
g
500
10
10
End sample input file

 Note:  The options are characters.
        They are the first character on their line.
        Choices are xx, yy, dd, or [gbre][gbre]

 Output: Distrfile contains the distribution in the following format
                    line 0              n q0 q1
                    line i+1            i D[i].s D[i].t for 0 <= i < n,
         where vector component i is allocated to 
         processor (D[i].s, D[i].t), with 0 <= D[i].s < q0
                                    and  0 <= D[i].t < q1. }
         


;const maxn  =  161000
;      maxq  =     100
;      m1 = 259200 

;type pair   = record s,t : 0..maxq end
;type distr= array[0..maxn]  of pair 
;type vec = array[0..maxn]  of integer
;type boolvec = array[0..maxn]  of boolean

;var inpfile, distrfile: text

{WARNING you have to plug in your own random number generator instead of ran1}
function ran1(var idum: integer): real;

;procedure randomperm(var idum: integer; n: integer; var sigma: vec)
  ;var i, range, draw, tmp: integer
  ;    ran: real
  ;    IsImage: boolvec
  ;    perm: boolean
  ;begin
      for i:=0 to n-1 do sigma[i]:= i
     ;for range:=n downto 2 do 
         begin
            ;ran:= ran1(idum)
            ;draw:= trunc(ran*range)
            ;if draw>=range then draw:= range-1
            {swap sigma[draw],sigma[range-1] }
           ;tmp:=sigma[draw]
           ;sigma[draw]:=sigma[range-1]
           ;sigma[range-1]:= tmp
         end
     ;perm:= true
     ;for i:=0 to n-1 do IsImage[i]:= false
     ;for i:=0 to n-1 do
        begin
           if (sigma[i]<0) or (sigma[i]>=n) then perm:= false
          ;IsImage[sigma[i]]:= true
        end
     ;for i:=0 to n-1 do 
         if not IsImage[i] then perm:= false
     ;if not perm then 
         writeln(output, ' *** Error: in random permutation')
   end {randomperm}

;procedure equalrandomdistr(var idum: integer; var D0: vec; n,q: integer)
  ;var i: integer
  ;    sigma: vec
  ;begin
      randomperm(idum,n,sigma)
     ;for i:=0 to n-1 do D0[i]:= sigma[i] mod q
   end {equalrandomdistr}

;procedure writedistr(D: distr; n,q0,q1: integer)
  ;var i: integer
  ;begin
      writeln(distrfile,n,q0,q1)
     ;for i:=0 to n-1 do writeln(distrfile,i,D[i].s,D[i].t)
   end {writedistr}

;procedure gbredistr(var idum: integer; ch: char; q, n: integer;
                     var D: vec)
   ;var L0, L1, R, i: integer
   ;    ran: real
   ;begin
       case ch of
       'g': for i:= 0 to n-1 do D[i]:= i mod q;
       'b': begin
               L0:= n div q
              ;R:= n mod q
              ;if R=0 then L1:= L0 else L1:= L0 + 1
              ;for i:= 0 to n-1 do
                 if i<R*L1 then D[i]:= i div L1
                           else D[i]:= R + ((i-R*L1) div L0)
            end;
       'r': for i:=0 to n-1 do
               begin
                 ;ran:= ran1(idum)
                 ;D[i]:= trunc(ran*q)
                 ;if D[i]>=q then D[i]:= q-1
               end;
       'e': equalrandomdistr(idum,D,n,q)
       end
    end {gbredistr}

;var D: distr 
;    D0,D1: vec
;    i,n,q0,q1,p,idum,proc,R,L0,L1: integer
;    error:integer
;    ch0, ch1: char

;begin {main}

    {******* Read matrix and options *******}
    idum:= - (wallclock mod m1) -1

    ;open(inpfile,'^1','old',error)
    ;if error<>0 then
        writeln(output,error,' *** Error: opening input file')
    ;reset(inpfile)

    ;open(distrfile,'^2','new',error)
    ;if error<>0 then
        writeln(output,error,' *** Error: opening distribution file')

    ;readln(inpfile,ch0)
    ;readln(inpfile,ch1)
    ;readln(inpfile,n)
    ;if n < 1 then writeln(output, ' *** Error: n < 1')
    ;readln(inpfile,q0)
    ;if q0 < 1 then writeln(output, ' *** Error: q0 < 1')
    ;readln(inpfile,q1)
    ;if q1 < 1 then writeln(output, ' *** Error: q1 < 1')
    ;p:=q0*q1

    {******* Create distribution *******}
    ;case ch0 of
    'x': { Diagonal by grids }
        begin
          for i:= 0 to n-1 do
              begin
                 proc:=  i mod p
                ;D0[i]:= proc mod q0
                ;D1[i]:= proc div q0
              end
        end;
    'y': { Diagonal by blocks }
        begin
           L0:= n div p
          ;R:= n mod p
          ;if R=0 then L1:= L0 else L1:= L0 + 1
          ;for i:= 0 to n-1 do
              begin
                 if i<R*L1 then proc:= i div L1
                           else proc:= R + ((i-R*L1) div L0)
                ;D0[i]:= proc mod q0
                ;D1[i]:= proc div q0
              end
        end;
     'd': { Diagonal randomly in equal sets }
        begin
          equalrandomdistr(idum,D0,n,p)
         ;for i:=0 to n-1 do
            begin
               proc:= D0[i]
              ;D0[i]:= proc mod q0
              ;D1[i]:= proc div q0
            end
        end;
     'g','b','r','e': { Grid, Block, Random, Equal-random}
        begin
           gbredistr(idum,ch0,q0,n,D0)
          ;gbredistr(idum,ch1,q1,n,D1)
        end
    end

    {******* Write distribution *******}
    ;for i:= 0 to n-1 do
      begin
         D[i].s:= D0[i]
        ;D[i].t:= D1[i]
      end
    ;writedistr(D,n,q0,q1) 
    ;close(inpfile)
    ;close(distrfile)
end.
