program mkdmd(input,output)

{This program generates a distribution
which corresponds to a molecular dynamics  matrix
    md.nparticles.cutoffinv   or
    mdr.nparticles.cutoffinv.densinv
of the report R. H. Bisseling and W. F. McColl,
Scientific Computing on Bulk Synchronous Parallel Architectures,
Preprint 836, Department of Mathematics, Utrecht University,
dec. 1993.

Usage: mkd inpfile matfile distrfile

  This program creates a Cartesian distribution
  of a vector of length n 
  over a q0 by q1 processor network,
  using an orthogonal domain decomposition scheme.
  The vector components represent particles
  in a d-dimensional molecular dynamics universe.

 Input: three filenames as arguments
        inpfile   = file containing
                    line 0     d = number of dimensions of
                                   the universe (must be 3)
                    line 1     P[r], 0 <= r < d 
                               P[r] = number of partitions of dimension r
                    line 2     q0 = number of processor rows
                    line 3     q1 = number of processor columns
                    where p = q0 * q1 = P[0]*P[1]*...*P[d-1]
                            = total number of processors
        matfile    = file containing the matrix in the following format
                    S n n              
                    (i j v)*
                    -1
                    line i after "-1": i x[i,r], 0 <= r < d,
                    for 0 <= i < n,
                    where n = matrix size
                          i = row index of a matrix element
                          j = column index of a matrix element
                          v = value of a matrix element
                          x[i,r] = r-th coordinate of particle i
                          (0 <= x[i,r] <=1)

        distrfile = new distribution file to be created 

Sample input file:
3
10 5 2
10
10
End sample input file


 Output: Distrfile contains the distribution in the following format
                    line 0              n q0 q1
                    line i+1            i D[i].s D[i].t for 0 <= i < n,
         where vector component i is allocated to 
         processor (D[i].s, D[i].t), with 0 <= D[i].s < q0
                                    and  0 <= D[i].t < q1. }
         


;const maxnz = 1440000
;      maxn  =  160000
;      maxq  =     100
;      maxd  =       3

;type patpair  = record i,j : 0..maxn end
;type pair   = record s,t : 0..maxq end
;type sprspat = array[0..maxnz] of patpair
;type distr= array[0..maxn]  of pair 
;type particle = array[0..maxd] of real 
;type particlemap = array[0..maxd] of integer

;var inpfile, matfile, distrfile: text

;procedure readpat(var A: sprspat; var n,nz: integer)
  ;var x,m : integer
  ;    ch  : char
  ;    val : real
  ;begin
      ch:=chr(20)
     ;while ch<>'S' do read(matfile,ch)
     ;read(matfile,m,n)
     ; if m<>n then
         writeln(output, '*** Error: matrix should be square')
     ;nz:=0
     ;read(matfile,x)
     ;while x<>-1 do
      begin
         A[nz].i:=x
        ;read(matfile,A[nz].j,val,x)
        ;nz:=nz+1
      end
  end {readpat}

;function Map (x: real; px: integer): integer
  ;var dx: real
  ;    pp: integer
  ;begin
      if x >= 0.0 then dx := x - trunc(x)
                  else dx:=  x - trunc(x) + 1
     ;if dx = 1.0 then dx:= 0.0
      { 0 <= dx < 1 }
     ;pp:= trunc(dx*px)
     ;if pp < 0 then Map:= 0
      else if pp >= px then Map:= px-1
      else Map:= pp
     { 0 <= Map < px }
   end {Map}

;function Pid(d: integer; var MapX,P: particlemap): integer
  { 0 <= MapX[r] < P[r], for 0 <= r < d }
  ;var sum, prod, r: integer
  ;begin
      sum:= 0
     ;prod:= 1 
     ;for r:= 0 to d-1 do
         begin
            sum:= sum + MapX[r]*prod
           ;prod:= prod*P[r]
         end
     ;Pid:= sum
   end {Pid}

;procedure writedistr(D: distr; n,q0,q1: integer)
  ;var i: integer
  ;begin
      writeln(distrfile,n,q0,q1)
     ;for i:=0 to n-1 do writeln(distrfile,i,D[i].s,D[i].t)
   end {writedistr}

;var A: sprspat
;    D: distr 
;    X: particle
;    MapX,P: particlemap
;    i,ii,n,nz,r,d,q0,q1,p,pp,proc: integer
;    error:integer

;begin {main}

    {******* Read matrix *******}

    ;open(inpfile,'^1','old',error)
    ;if error<>0 then
        writeln(output,error,' *** Error: opening input file')
    ;reset(inpfile)

    ;open(matfile,'^2','old',error)
    ;if error<>0 then
        writeln(output,error,'*** Error: opening matrix file')

    ;open(distrfile,'^3','new',error)
    ;if error<>0 then
        writeln(output,error,' *** Error: opening distribution file')

    ;readln(inpfile,d)
    ;if d < 1 then writeln(output, ' *** Error: d < 1')
    ;for r:= 0 to d-1 do read(inpfile,P[r])
    ;readln(inpfile)
    ;readln(inpfile,q0)
    ;if q0 < 1 then writeln(output, ' *** Error: q0 < 1')
    ;readln(inpfile,q1)
    ;if q1 < 1 then writeln(output, ' *** Error: q1 < 1')
    ;p:=q0*q1
    ;pp:= 1
    ;for r:= 0 to d-1 do pp:= pp*P[r]
    ;if p <> pp then
        writeln(output, ' *** Error: q0*q1 <> P[0]*...*P[d-1]')

    {******* Read particle coordinates and create distribution *******}
    ;readpat(A,n,nz)
    ;for i:= 0 to n-1 do
        begin
           read(matfile,ii)
          ;if i <> ii then
              writeln(output, ' *** Error: reading particle coordinates')
          ;for r:= 0 to d-1 do read(matfile,X[r]) { X[r] = x[i,r] }
          ;readln(matfile)
          ;for r:= 0 to d-1 do MapX[r]:= Map(X[r],P[r])
          ;proc:= Pid(d,MapX,P) 
          ;D[i].s:= proc mod q0
          ;D[i].t:= proc div q0
        end

    {******* Write distribution *******}
    ;writedistr(D,n,q0,q1) 
    ;close(inpfile)
    ;close(matfile)
    ;close(distrfile)
end.
